use crate::*;

pub fn cell_diff_4_vertical_small(prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	const CELLS_X: usize = WIDTH / 4;
	const CELLS_Y: usize = HEIGHT / 4;

	let mut modified_cells = [[false; CELLS_Y]; CELLS_X];

	let (mut cell_runs, modified_cells_flat) =
		rle_255_decode(&unpack_nybbles(encoded), CELLS_X * CELLS_Y);

	if cell_runs % 2 == 1 {
		cell_runs += 1;
	}
	*reader += cell_runs / 2;
	let encoded = &encoded[(cell_runs / 2)..];

	let mut changed_cell_count = 0;
	let mut i = 0;
	for cellx in 0..CELLS_X {
		for celly in 0..CELLS_Y {
			let cell = modified_cells_flat[i] == 1;
			modified_cells[cellx][celly] = cell;
			if cell {
				changed_cell_count += 1;
			}
			i += 1;
		}
	}

	let mut frame = *prev_frame;

	let changed_pixel_count = changed_cell_count * 4 * 4;
	let (runs, new_pixels) = rle_255_decode(encoded, changed_pixel_count);
	*reader += runs;

	let mut index = 0;

	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			let cellx = x / 4;
			let celly = y / 4;
			let is_changed = modified_cells[cellx][celly];
			if is_changed {
				frame[x][y] = new_pixels[index];
				index += 1;
			}
		}
	}

	frame
}

pub fn cell_diff_8_vertical(prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	let bitmap = {
		*reader += 3;
		let mut bitmap = (encoded[0] as u32) << 16;
		bitmap |= (encoded[1] as u32) << 8;
		bitmap | encoded[2] as u32
	};
	let encoded = &encoded[3..];

	let mut frame = *prev_frame;
	let changed_pixel_count = bitmap.count_ones() as usize * 8 * 8;
	let (runs, new_pixels) = rle_255_decode(encoded, changed_pixel_count);
	*reader += runs;

	let cells_x = WIDTH / 8;
	let mut index = 0;

	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			let cellx = x / 8;
			let celly = y / 8;
			let is_changed = (bitmap >> (cellx + cells_x * celly)) & 1;
			if is_changed == 1 {
				frame[x][y] = new_pixels[index];
				index += 1;
			}
		}
	}

	frame
}

pub fn bg_strips_horizontal16(_prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	*reader += 1;
	let bg = encoded[0] >> 7;
	let fg = 1 - bg;
	let mut frame = [[bg; HEIGHT]; WIDTH];
	let count = (encoded[0] & 0x7f) as usize;
	*reader += count * 2;
	for i in 0..count {
		let upper = (encoded[i * 2 + 1] as u16) << 8;
		let lower = encoded[i * 2 + 2] as u16;
		let y_x_w: u16 = upper | lower;
		let y = (y_x_w >> 11) as usize;
		let x_start = ((y_x_w >> 5) & 0x3f) as usize;
		let w = (y_x_w & 0x1f) as usize;
		for x in x_start..(x_start + w) {
			frame[x][y] = fg;
		}
	}
	frame
}

pub fn bg_strips_horizontal24(_prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	*reader += 2;
	let bg = encoded[0] >> 7;
	let fg = 1 - bg;
	let mut frame = [[bg; HEIGHT]; WIDTH];
	let count_upper = (encoded[0] & 0x7f) as usize;
	let count = (count_upper << 8) | encoded[1] as usize;
	*reader += count * 3;
	for i in 0..count {
		let y = encoded[i * 3 + 2] as usize;
		let x_lower = encoded[i * 3 + 3];
		let x_width = encoded[i * 3 + 4];
		let x_upper = ((x_width & 0x80) as usize) << 1;
		let x_start = x_lower as usize | x_upper;
		let width = (x_width & 0x7f) as usize;
		for x in x_start..(x_start + width) {
			frame[x][y] = fg;
		}
	}
	frame
}


pub fn rle_diff_horizontal(prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	let (runs, decoded) = rle_255_decode(encoded, FRAME_SIZE);
	*reader += runs;
	let mut frame = *prev_frame;
	let mut i = 0;
	for y in 0..HEIGHT {
		for x in 0..WIDTH {
			frame[x][y] ^= decoded[i];
			i += 1;
		}
	}
	frame
}

pub fn rle_diff_vertical(prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	let (runs, decoded) = rle_255_decode(encoded, FRAME_SIZE);
	*reader += runs;
	let mut frame = *prev_frame;
	let mut dbg_frame = FRAME_0;
	let mut i = 0;
	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			frame[x][y] ^= decoded[i];
			frame[x][y] &= 1;
			dbg_frame[x][y] ^= decoded[i];
			dbg_frame[x][y] &= 1;
			i += 1;
		}
	}
	frame
}

pub fn rle_horizontal(_prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	let (runs, pixels) = rle_255_decode(encoded, FRAME_SIZE);
	*reader += runs;
	let mut frame = FRAME_0;
	let mut i = 0;
	for y in 0..HEIGHT {
		for x in 0..WIDTH {
			frame[x][y] = pixels[i];
			i += 1;
		}
	}
	frame
}

pub fn rle_vertical(_prev_frame: &Frame, encoded: &[u8], reader: &mut usize) -> Frame {
	let (runs, pixels) = rle_255_decode(encoded, FRAME_SIZE);
	*reader += runs;
	let mut frame = FRAME_0;
	let mut i = 0;
	for x in 0..WIDTH {
		for y in 0..HEIGHT {
			frame[x][y] = pixels[i];
			i += 1;
		}
	}
	frame
}

pub fn fill_white(_prev_frame: &Frame, _encoded: &[u8], _reader: &mut usize) -> Frame {
	FRAME_1
}

pub fn fill_black(_prev_frame: &Frame, _encoded: &[u8], _reader: &mut usize) -> Frame {
	FRAME_0
}
